@@ -1,19 +1,37 @@
 from io import BytesIO
+from json import loads as json_loads
 from pprint import pformat
 from unittest.mock import call, patch, mock_open
 
 from pytest import fixture, raises
 
-from pymw import API, LoginError, APIError
-
+from pymw import API, LoginError, APIError, _api
+# noinspection PyProtectedMember
+from pymw._api import get_lgname_lgpass, load_config
 
 url = 'https://www.mediawiki.org/w/api.php'
 api = API(url)
 
 
+TEXT_CONFIG = '''
+{
+    "https://www.mediawiki.org/w/api.php": {
+        "U@T": {"BotPassword": "BP"}
+    },
+    "https://*.wikipedia.org/w/api.php": {
+        "U": {"BotPassword": "P"}
+    }
+}
+'''
+CONFIG = json_loads(TEXT_CONFIG)
+json_open_mock = mock_open(read_data=TEXT_CONFIG)
+patch.object(_api, 'CONFIG', CONFIG).start()
+
+
 @fixture
 def cleared_api():
-    api.clear_cache()
+    api.tokens.clear()
+    api._user = None
     return api
 
 
@@ -32,22 +50,30 @@ class FakeResp:
         return self._json
 
 
-def patch_post(obj, attr, call_returns, ):
+def patch_post(obj, attr, call_returns, api_post):
     i = -2
 
     def side_effect(*args, **kwargs):
         nonlocal i
         i += 2
-        if (call := call_returns[i]) is not any:
-            assert args == call.args
-            assert kwargs == call.kwargs
+        if (expected := call_returns[i]) is not any:
+            actual_pop = kwargs.pop
+            call_pop = (expected_kwargs := expected.kwargs).pop
+            if api_post == 1:  # API.post
+                assert args == expected.args
+            else:  # session.post
+                assert not args
+                assert expected.args[0] == actual_pop('data')
+            assert call_pop('files', None) == actual_pop('files', None)
+            assert call_pop('params', None) == actual_pop('params', None)
+            assert not kwargs and not expected_kwargs
         return call_returns[i + 1]
 
     return patch.object(obj, attr, side_effect=side_effect)
 
 
 def api_post_patch(*call_returns):
-    return patch_post(API, 'post', call_returns)
+    return patch_post(API, 'post', call_returns, True)
 
 
 def session_post_patch(*call_header_returns):
@@ -65,7 +91,7 @@ def session_post_patch(*call_header_returns):
             response = FakeResp(json=headers_or_json)
             call_responses += (call, response)
             call = json_or_call
-    return patch_post(api.session, 'request', call_responses)
+    return patch_post(api, '_post', call_responses, False)
 
 
 @api_post_patch(
@@ -117,15 +143,15 @@ def test_recentchanges(_):
 @patch('pymw._api.sleep', fake_sleep)
 @patch('pymw._api.warning')
 @session_post_patch(
-    call('POST', url, data={
+    call({
         'action': 'query', 'errorformat': 'plaintext', 'format': 'json',
         'formatversion': '2', 'maxlag': 5, 'meta': 'tokens', 'type': 'watch'
-    }, files=None),
-    {'retry-after': '5'}, {'errors': [{'code': 'maxlag', 'text': 'Waiting for 10.64.16.7: 0.80593395233154 seconds lagged.', 'data': {'host': '10.64.16.7', 'lag': 0.805933952331543, 'type': 'db'}, 'module': 'main'}], 'docref': 'See https://www.mediawiki.org/w/api.php for API usage. Subscribe to the mediawiki-api-announce mailing list at &lt;https://lists.wikimedia.org/mailman/listinfo/mediawiki-api-announce&gt; for notice of API deprecations and breaking changes.', 'servedby': 'mw1225'},
-    call('POST', url, data={
+    }),
+    {'retry-after': '5'}, {'errors': [{'code': 'maxlag', 'text': 'Waiting for 10.64.16.7: 0.80593395233154 seconds lagged.', 'data': {'host': '10.64.16.7', 'lag': 0.805933952331543, 'type': 'db'}, 'module': 'main'}], 'docref': ..., 'servedby': 'mw1225'},
+    call({
         'meta': 'tokens', 'type': 'watch', 'action': 'query', 'format': 'json',
         'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5
-    }, files=None),
+    }),
     {}, {'batchcomplete': True, 'query': {'tokens': {'watchtoken': '+\\'}}})
 def test_maxlag(_, warning_mock, cleared_api):
     tokens = cleared_api.query_meta('tokens', type='watch')
@@ -199,32 +225,9 @@ def test_context_manager():
     close_mock.assert_called_once_with()
 
 
-@session_post_patch(
-    any, {}, {'batchcomplete': True, 'query': {'tokens': {'patroltoken': '+\\'}}},
-    call('POST', url, data={
-        'revid': 27040231, 'action': 'patrol', 'token': '+\\', 'format':
-            'json', 'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5
-    }, files=None), {}, {'errors': [{'code': 'permissiondenied', 'text': 'T', 'module': 'patrol'}], 'docref': 'D', 'servedby': 'mw1233'})
-def test_patrol_not_logged_in(_, cleared_api):
-    try:
-        cleared_api.patrol(revid=27040231)
-    except APIError:
-        pass
-    else:  # pragma: nocover
-        raise AssertionError('APIError was not raised')
-
-
-@api_post_patch(
-    call({'action': 'patrol', 'token': 'T', 'revid': 1}),
-    {'patrol': {'rcid': 1, 'ns': 4, 'title': 'T'}})
-def test_patrol(post_mock):
-    api.tokens['patrol'] = 'T'
-    api.patrol(revid=1)
-    post_mock.assert_called_once()
-
-
-@session_post_patch(any, {}, {'errors': [{'code': 'badtoken', 'text': 'Invalid CSRF token.', 'module': 'patrol'}], 'docref': 'D', 'servedby': 'mw1279'})
+@session_post_patch(any, {}, {'errors': [{'code': 'badtoken', 'text': 'Invalid CSRF token.', 'module': 'patrol'}], 'docref': ..., 'servedby': 'mw1279'})
 def test_bad_patrol_token(_):
+    api._user = 'x'
     api.tokens['patrol'] = 'T'
     try:
         api.patrol(revid=1)
@@ -300,11 +303,10 @@ def test_revisions_mode2_no_rvlimit(post_mock):  # auto set rvlimit
 
 
 @api_post_patch(
-    call({'action': 'upload', 'token': 'T', 'filename': 'FN.jpg'}, files={'file': ('FN.jpg', NotImplemented)}),
+    call({'action': 'upload', 'filename': 'FN.jpg'}, files={'file': ('FN.jpg', NotImplemented)}),
     {'upload': {'result': 'Warning', 'warnings': {'exists': 'Test.jpg', 'nochange': {'timestamp': '2020-07-04T07:29:07Z'}}, 'filekey': 'tampered.y27er1.18.jpg', 'sessionkey': 'tampered.y27er1.18.jpg'}})
 def test_upload_file(post_mock):
-    api._assert_user = 'James Bond'
-    api.tokens['csrf'] = 'T'
+    api._user = 'James Bond'
     api.upload_file(file=NotImplemented, filename='FN.jpg')
     post_mock.assert_called_once()
 
@@ -312,7 +314,7 @@ def test_upload_file(post_mock):
 @patch.object(API, 'login')
 def test_upload_file_auto_login(login_mock):
     login_mock.side_effect = NotImplementedError
-    api._assert_user = None
+    api._user = None
     with raises(NotImplementedError):
         api.upload_file(file=NotImplemented, filename='FN.jpg')
     login_mock.assert_called_once_with()
@@ -324,21 +326,18 @@ bio1 = BytesIO(b'1')
 
 @api_post_patch(
     call(
-        {'action': 'upload', 'token': 'T', 'stash': 1, 'offset': 0, 'filename': 'F.jpg', 'filesize': 5039, 'ignorewarnings': True},
+        {'action': 'upload', 'stash': 1, 'offset': 0, 'filename': 'F.jpg', 'filesize': 5039, 'ignorewarnings': True},
         files={'chunk': ('F.jpg', bio0)}),
     {'upload': {'warnings': {'duplicate-archive': 'F.jpg'}, 'result': 'Continue', 'offset': 3000, 'filekey': 'K'}},
     call(
-        {'action': 'upload', 'token': 'T', 'stash': 1, 'offset': 3000, 'filename': 'F.jpg', 'filesize': 5039, 'ignorewarnings': True, 'filekey': 'K'},
+        {'action': 'upload', 'stash': 1, 'offset': 3000, 'filename': 'F.jpg', 'filesize': 5039, 'ignorewarnings': True, 'filekey': 'K'},
         files={'chunk': ('F.jpg', bio1)}),
     {'upload': {'filekey': 'K.jpg', 'imageinfo': {'CENSORED': ...}, 'result': 'Success', 'warnings': {'duplicate-archive': 'T.jpg'}}},
     call(
-        {'action': 'upload', 'token': 'T', 'filename': 'F.jpg', 'ignorewarnings': True, 'filekey': 'K.jpg'},
-        files=None),
+        {'action': 'upload', 'filename': 'F.jpg', 'ignorewarnings': True, 'filekey': 'K.jpg'}),
     {'upload': {'filename': 'F.jpg', 'imageinfo': {'CENSORED': ...}, 'result': 'Success'}})
 def test_upload_chunks(_):
-    api._assert_user = 'U'
-    api.tokens['csrf'] = 'T'
-
+    api._user = 'U'
     result = api.upload_chunks(
         chunks=(b for b in (bio0, bio1)),
         filename='F.jpg',
@@ -347,64 +346,47 @@ def test_upload_chunks(_):
     assert result == {'filename': 'F.jpg', 'imageinfo': {'CENSORED': ...}, 'result': 'Success'}
 
 
-pymw_toml = '''
-version = 1
-
-['https://www.mediawiki.org/w/api.php'.login]
-'username@toolname' = 'bot_password'
-'''
-pymw_toml_mock = mock_open(read_data=pymw_toml)
-
-
-@patch('pymw._api.Path.open', pymw_toml_mock)
-@api_post_patch(any, {}, any, {})
-def test_login_config(post_mock):
-    post_call_data = {
-        'action': 'login', 'lgname': 'username@toolname',
-        'lgpassword': 'bot_password', 'lgtoken': 'LOGIN_TOKEN'}
-    api.tokens['login'] = 'LOGIN_TOKEN'
-    with raises(KeyError):  # because of invalid api_post_patch response
-        api.login()  # without username
-    post_mock.assert_called_once_with(post_call_data)
-    with raises(KeyError):  # again, because of invalid api_post_patch response
-        api.login('username@toolname')  # without username
-    # note that assert_called_with only checks the last call
-    post_mock.assert_called_with(post_call_data)
-    pymw_toml_mock.assert_called_once()
+@patch('pymw._api.Path.open', json_open_mock)
+@patch.object(_api, 'CONFIG', None)
+def test_login_config():
+    _api.load_config()
+    assert _api.CONFIG == CONFIG
+    json_open_mock.assert_called_once()
 
 
 @session_post_patch(any, {}, {})
 def test_assert_login(post_mock):
-    api._assert_user = 'USER'
+    api._user = 'USER'
     api.post({})
     assert post_mock.mock_calls[0].kwargs['data']['assertuser'] == 'USER'
 
 
-@patch('pymw._api.Path.open', pymw_toml_mock)
 @session_post_patch(
-    call('POST', url, data={
+    call({
         'notfilter': '!read', 'meta': 'notifications', 'action': 'query',
         'errorformat': 'plaintext', 'format': 'json', 'formatversion': '2',
-        'maxlag': 5}, files=None),
-    {'errors': [{'code': 'login-required', 'text': 'You must be logged in.', 'module': 'query+notifications'}], 'docref': '', 'servedby': 'mw1341'},
-    call('POST', url, data={
+        'maxlag': 5}),
+    {'errors': [{'code': 'login-required', 'text': 'You must be logged in.', 'module': 'query+notifications'}], 'docref': ..., 'servedby': 'mw1341'},
+    call({
         'type': 'login', 'meta': 'tokens', 'action': 'query', 'format': 'json',
-        'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5}, files=None),
+        'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5}),
     {'batchcomplete': True, 'query': {'tokens': {'logintoken': 'T1'}}},
-    call('POST', url, data={
-        'action': 'login', 'lgname': 'username@toolname', 'lgpassword':
-            'bot_password', 'lgtoken': 'T1', 'format': 'json', 'formatversion':
-            '2', 'errorformat': 'plaintext', 'maxlag': 5}, files=None),
+    call({
+        'action': 'login', 'lgname': 'U@T', 'lgpassword':
+            'BP', 'lgtoken': 'T1', 'format': 'json', 'formatversion':
+            '2', 'errorformat': 'plaintext', 'maxlag': 5}),
     {'login': {'result': 'Success', 'lguserid': 1, 'lgusername': 'username'}},
-    call('POST', url, data={
+    call({
         'notfilter': '!read', 'meta': 'notifications', 'action': 'query',
         'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext',
-        'maxlag': 5, 'assertuser': 'username'}, files=None),
+        'maxlag': 5, 'assertuser': 'username'}),
     {'batchcomplete': True, 'query': {'notifications': {'list': [], 'continue': None}}},
 )
 def test_handle_login_required(_, cleared_api):
-    r = api.query_meta('notifications', notfilter='!read')
+    assert cleared_api.user is None
+    r = cleared_api.query_meta('notifications', notfilter='!read')
     assert r == {'list': [], 'continue': None}
+    assert cleared_api.user == 'username'
 
 
 @api_post_patch(
@@ -526,3 +508,162 @@ page3 = {'ns': 0, 'pageid': 1844356, 'revisions': [{
 def test_page_in_last_batch(_):
     page = next(api.query_prop('revisions'))
     assert page is page3
+
+
+@api_post_patch(
+    any, {
+        'continue': {'llcontinue': '288753|en', 'continue': '||'},
+        'query': {'pages': [{
+            'pageid': 288753, 'ns': 0, 'title': 'شیران',
+            'langlinks': [
+                {'lang': 'ar', 'title': 'شيران'},
+                {'lang': 'arz', 'title': 'شيران'}]}]}},
+    any, {
+        'continue': {'llcontinue': '288753|zh-min-nan', 'continue': '||'},
+        'query': {'pages': [{
+            'pageid': 288753, 'ns': 0, 'title': 'شیران', 'langlinks': [
+                {'lang': 'en', 'title': 'Shiran, Ardabil'},
+                {'lang': 'ms', 'title': 'Shiran, Ardabil'}]}]}},
+    any, {'batchcomplete': True, 'query': {'pages': [
+        {'pageid': 288753, 'ns': 0, 'title': 'شیران', 'langlinks': [
+            {'lang': 'zh-min-nan', 'title': 'Shiran (Ardabil)'}]}]}})
+def test_batch_prop_extend(_):
+    assert next(api.query_prop(
+        'langlinks', llprop='', lllimit=2, titles='شیران')
+    ) == {
+        'pageid': 288753, 'ns': 0, 'title': 'شیران', 'langlinks': [
+            {'lang': 'ar', 'title': 'شيران'},
+            {'lang': 'arz', 'title': 'شيران'},
+            {'lang': 'en', 'title': 'Shiran, Ardabil'},
+            {'lang': 'ms', 'title': 'Shiran, Ardabil'},
+            {'lang': 'zh-min-nan', 'title': 'Shiran (Ardabil)'}]}
+
+
+def test_url_property():
+    assert api.url == url
+    with raises(AttributeError):  # can't set attribute
+        # noinspection PyPropertyAccess
+        api.url = ''
+
+
+def test_repr():
+    assert repr(api) == f'API({repr(url)})'
+
+
+def test_glob_pattern_load_lgname_lgpass():
+    assert get_lgname_lgpass('https://en.wikipedia.org/w/api.php') \
+            == get_lgname_lgpass('https://en.wikipedia.org/w/api.php', 'U') \
+            == ('U', 'P')
+
+
+watch_response = {'batchcomplete': True, 'watch': [{'title': '0', 'ns': 0, 'unwatched': True}, {'title': '1', 'ns': 0, 'unwatched': True}]}
+
+
+@session_post_patch(
+    call({
+        'action': 'watch', 'titles': '0|1', 'format': 'json',
+        'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5,
+        'token': '+\\', 'unwatch': True}),
+    {'errors': [{
+        'code': 'notloggedin',
+        'text': 'Please log in to view or edit items on your watchlist.',
+        'module': 'watch'}], 'docref': ..., 'servedby': 'mw1342'},
+    call({
+        'type': 'login', 'meta': 'tokens', 'action': 'query', 'format': 'json',
+        'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5
+    }), {'batchcomplete': True, 'query': {'tokens': {
+        'logintoken': 'LGT+\\'}}},
+    call({
+        'action': 'login', 'lgname': 'U@T',
+        'lgpassword': 'BP', 'lgtoken': 'LGT+\\',
+        'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext',
+        'maxlag': 5}),
+    {'login': {'result': 'Success', 'lguserid': 1, 'lgusername': 'username'}},
+    call({
+        'type': 'watch', 'meta': 'tokens', 'action': 'query', 'format': 'json',
+        'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5,
+        'assertuser': 'username'}),
+    {'batchcomplete': True, 'query': {'tokens': {'watchtoken': 'LIWT+\\'}}},
+    call({
+        'action': 'watch', 'titles': '0|1', 'format': 'json',
+        'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5,
+        'token': 'LIWT+\\',
+        'unwatch': True, 'assertuser': 'username'}),
+    watch_response,)
+# assume that watch is not marked as a login-required action
+@patch.dict('pymw._api.LOGIN_REQUIRED_ACTIONS', clear=True)
+def test_notloggedin_error(_post_mock, cleared_api):
+    cleared_api.tokens['watch'] = '+\\'
+    r = cleared_api.post({
+        'action': 'watch', 'titles': '0|1', 'format': 'json',
+        'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5,
+        'unwatch': True})
+    assert r is watch_response
+
+
+spamblacklist_ok = {'spamblacklist': {'result': 'ok'}}
+
+
+@session_post_patch(
+    any, {'errors': [{
+        'code': 'toomanyvalues',
+        'text': 'Too many values supplied for parameter "url". The limit is 50.',
+        'data': {'limit': 50, 'lowlimit': 50, 'highlimit': 500},
+        'module': 'spamblacklist'}], 'docref': ..., 'servedby': 'mw1356'},
+    call({
+        'action': 'spamblacklist',
+        'url': '0|1|2|3|4|5|6|7|8|9|10|11|12|13|14|15|16|17|18|19|20|21|22|23'
+               '|24|25|26|27|28|29|30|31|32|33|34|35|36|37|38|39|40|41|42|43|44'
+               '|45|46|47|48|49',
+        'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext',
+        'maxlag': 5}), spamblacklist_ok,
+    call({
+        'action': 'spamblacklist', 'url': '50', 'format': 'json',
+        'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5}),
+    spamblacklist_ok)
+@patch('pymw._api.warning')
+def test_handle_toomanyvalues_in_post_and_continue(warning, _, cleared_api):
+    for r in cleared_api.post_and_continue({
+            'action': 'spamblacklist',
+            'url': '0|1|2|3|4|5|6|7|8|9|10|11|12|13|14|15|16|17|18|19|20|21|22'
+                   '|23|24|25|26|27|28|29|30|31|32|33|34|35|36|37|38|39|40|41'
+                   '|42|43|44|45|46|47|48|49|50',
+            'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext',
+            'maxlag': 5}):
+        assert r is spamblacklist_ok
+    warning.assert_called_once_with(
+        '`toomanyvalues` error occurred; '
+        'trying to split `url` into several API calls.\n'
+        'NOTE: sometimes doing this does not make sense.')
+
+
+@session_post_patch(call({
+    'action': 'paraminfo', 'modules': 'query+info|query+categorymembers',
+    'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext',
+    'maxlag': 5}), {})
+def test_iterable_values_to_str(_):
+    api.post({
+        'action': 'paraminfo',
+        'modules': ('query+info', 'query+categorymembers')})
+
+
+@session_post_patch(
+    call({'type': 'login', 'meta': 'tokens', 'action': 'query', 'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5}),
+    {'batchcomplete': True, 'query': {'tokens': {'logintoken': 1}}},
+    call({'action': 'login', 'lgname': 'U@T', 'lgpassword': 'BP', 'lgtoken': 1, 'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5}),
+    {'login': {'result': 'Success', 'lguserid': 1, 'lgusername': 'U'}},
+    call({'type': 'csrf', 'meta': 'tokens', 'action': 'query', 'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5, 'assertuser': 'U'}),
+    {'batchcomplete': True, 'query': {'tokens': {'csrftoken': 2}}},
+    call({'action': 'undelete', 'title': 'a', 'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5, 'token': 2, 'assertuser': 'U'}),
+    (expected_response := {'undelete': {'title': 'A', 'revisions': 1, 'fileversions': 0, 'reason': ''}}))
+def test_post_and_continue_with_limited_action_without_limited_param(_, cleared_api):
+    assert (expected_response,) == (
+        *cleared_api.post_and_continue({'action': 'undelete', 'title': 'a'}),)
+
+
+@session_post_patch(
+    call({'action': 'query', 'meta': 'userinfo', 'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5}),
+    (ui_response := {'batchcomplete': True, 'query': {'userinfo': {'id': 0, 'name': '31.14.154.1', 'anon': True}}}))
+def test_empty_limited_value(_, cleared_api):
+    for r in cleared_api.post_and_continue({'action': 'query', 'meta': 'userinfo', 'titles': ''}):
+        assert r is ui_response
